import asyncio
from concurrent.futures import ProcessPoolExecutor
import math
import dill
from types import FunctionType
import psutil
from typing import Any, Coroutine, Dict, List, Tuple, Union
from hedra.core.engines.types.common.base_result import BaseResult
from .synchronization import BatchedSemaphore



class BatchExecutor:

    def __init__(
        self, 
        max_workers: int = psutil.cpu_count(logical=False)
    ) -> None:
        self.max_workers = max_workers
        self.loop = asyncio.get_event_loop()
        self.sem = BatchedSemaphore(max_workers)
        self.pool = ProcessPoolExecutor(max_workers=max_workers)
        self.shutdown_task = None

    async def execute_batches(self, batched_stages: List[Tuple[int, List[Any]]], execution_task: FunctionType) -> List[Tuple[str, Any]]:

        return await asyncio.gather(*[
            asyncio.create_task(
                self._execute_stage(
                    stage_name,
                    assigned_workers_count,
                    execution_task,
                    configs
                )
            ) for stage_name, assigned_workers_count, configs in batched_stages
        ])


    async def _execute_stage(
        self, 
        stage_name: str,
        assigned_workers_count: int,
        execution_task: FunctionType,
        configs: List[List[Any]]
    ) -> Tuple[str, Any]:
        await self.sem.acquire(assigned_workers_count)
        stage_results = await self.execute_stage_batch(
            execution_task,
            configs
        )

        self.sem.release(assigned_workers_count)

        return (stage_name, stage_results)

    async def execute_stage_batch(
        self, 
        execution_task: FunctionType,
        configs: List[Any]
    ):
        results = await asyncio.gather(*[
            self.loop.run_in_executor(
                self.pool,
                execution_task,
                config
            ) for config in configs
        ])
        
        return results

    def partion_stage_batches(self, stages: List[Any], ) -> List[Tuple[str, Any, int]]:

        # How many batches do we have? For example -> 5 stages over 4
        # CPUs means 2 batches. The first batch will assign one stage to
        # each core. The second will assign all four cores to the remaing
        # one stage.    

        stages_count = len(stages)


        self.batch_sets_count = math.ceil(stages_count/self.max_workers)
        if self.batch_sets_count < 1:
            self.batch_sets_count = 1
        
        self.more_stages_than_cpus = stages_count/self.max_workers > 1

        if stages_count%self.max_workers > 0 and stages_count > self.max_workers:

            batch_size = self.max_workers
            workers_per_stage = int(self.max_workers/batch_size)
            batched_stages = [
                [
                    workers_per_stage for _ in range(batch_size)
                ] for _ in range(self.batch_sets_count - 1)
            ]

        else:
            batch_size = int(stages_count/self.batch_sets_count)
            workers_per_stage = int(self.max_workers/batch_size)
            batched_stages = [
                [
                    workers_per_stage for _ in range(batch_size)
                ] for _ in range(self.batch_sets_count)
            ]

        batches_count = len(batched_stages)

        # If we have a remainder batch - i.e. more stages than cores.
        last_batch = batched_stages[batches_count-1]

        last_batch_size = stages_count%self.max_workers
        if last_batch_size > 0 and self.more_stages_than_cpus:
            last_batch_workers = int(self.max_workers/last_batch_size)
            batched_stages.append([
                last_batch_workers for _ in range(last_batch_size)
            ])

        last_batch = batched_stages[self.batch_sets_count-1]
        last_batch_size = len(last_batch)
        last_batch_remainder = self.max_workers%last_batch_size


        if last_batch_remainder > 0:
            for idx in range(last_batch_remainder):
                last_batch[idx] += 1

        stage_idx = 0
        batches = []

        for batch in batched_stages:
            for stage_idx, stage_workers_count in enumerate(batch):

                stage_name, stage = stages[stage_idx]

                batches.append((
                    stage_name,
                    stage,
                    stage_workers_count
                ))

        return batches




class Mean:

    __slots__ = (
        'size',
        'previous_size',
        'mean',
        'delta',
        'delta_n'
    )

    def __init__(self):
        self.size = 0
        self.previous_size = 0
        self.mean = 0.0
        self.delta = 0.0
        self.delta_n = 0.0

    def update(self, new_value):
        self.previous_size = self.size
        self.size += 1
        self.delta = new_value - self.mean
        self.delta_n = self.delta / self.size
        self.mean += self.delta_n

        return self

    def get(self):
        return self.mean




import re
import uuid
from typing import List
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet


try:
    from hedra.reporting.types.statsd import StatsD
    from aio_statsd import GraphiteClient
    from .graphite_config import GraphiteConfig
    has_connector = True

except Exception:
    from hedra.reporting.types.empty import Empty as StatsD
    GraphiteClient = None
    GraphiteConfig = None
    has_connector = False


class Graphite(StatsD):

    def __init__(self, config: GraphiteConfig) -> None:
        super().__init__(config)

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self.connection = GraphiteClient(
            host=self.host,
            port=self.port
        )

        self.statsd_type = 'Graphite'

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to {self.statsd_type}')

        for event in events:
            self.connection.send_graphite(f'{event.name}_time', event.time)
            
            if event.success:
                self.connection.send_graphite(f'{event.name}_success', 1)
            
            else:
                self.connection.send_graphite(f'{event.name}_failed', 1)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to {self.statsd_type}')


    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to {self.statsd_type}')
        
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for field, value in metrics_set.common_stats.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metric - {metrics_set.name}:common:{field}')

                self.connection.send_graphite(
                    f'{metrics_set.name}_common_{field}', value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to {self.statsd_type}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to {self.statsd_type}')

        for metrics_set in metrics:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
            
                for metric_field, metric_value in group.stats.items():

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metric - {metrics_set.name}:{group_name}:{metric_field}')

                    self.connection.send_graphite(
                        f'{metrics_set.name}_{group_name}_{metric_field}', metric_value
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitted Metric - {metrics_set.name}:{group_name}:{metric_field}')

                for metric_field, metric_value in group.custom.items():

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metric - {metrics_set.name}:{group_name}:{metric_field}')

                    self.connection.send_graphite(
                        f'{metrics_set.name}_{group_name}_{metric_field}', metric_value
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to {self.statsd_type}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to {self.statsd_type}')

        for metrics_set in metrics_sets:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_group_name, group in metrics_set.custom_metrics.items():

                for field, value in group.items():

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metric - {metrics_set.name}:{custom_group_name}:{field}')

                    self.connection.send_graphite(
                        f'{metrics_set.name}_{custom_group_name}_{field}', value
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to {self.statsd_type}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to {self.statsd_type}')

        for metrics_set in metrics_sets:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                error_message = re.sub(
                    '[^0-9a-zA-Z]+', 
                    '_',
                    error.get(
                        'message'
                    ).lower()
                )

                self.connection.send_graphite(
                    f'{metrics_set.name}_error_{error_message}',
                    error.get('count')
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to {self.statsd_type}')


import asyncio
import functools
import json
import uuid
import psutil
from typing import List
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet
from hedra.reporting.types import ReporterTypes
from .aws_lambda_config import AWSLambdaConfig

try:
    import boto3
    has_connector=True
except Exception:
    boto3 = None
    has_connector=False


class AWSLambda:

    def __init__(self, config: AWSLambdaConfig) -> None:
        self.aws_access_key_id = config.aws_access_key_id
        self.aws_secret_access_key = config.aws_secret_access_key
        self.region_name = config.region_name

        self.events_lambda_name = config.events_lambda
        self.metrics_lambda_name = config.metrics_lambda 

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._client = None
        self._loop = asyncio.get_event_loop()
        self.session_uuid = str(uuid.uuid4())

        self.reporter_type = ReporterTypes.AWSLambda
        self.reporter_type_name = self.reporter_type.name.capitalize()
        self.metadata_string: str = None

        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Opening session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to AWS - Region: {self.region_name}')

        self._client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'lambda',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Successfully opened connection to AWS - Region: {self.region_name}')

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Lambda - {self.events_lambda_name}')

        for event in events:

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self._client.invoke,
                    FunctionName=self.events_lambda_name,
                    Payload=json.dumps(event.record)
                )
            )

        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Lambda - {self.events_lambda_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Lambda - {self.metrics_lambda_name}')

        for metrics_set in metrics_sets:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self._client.invoke,
                    FunctionName=self.metrics_lambda_name,
                        Payload=json.dumps({
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': 'common',
                            **metrics_set.common_stats
                        })
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Lambda - {self.metrics_lambda_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Lambda - {self.metrics_lambda_name}')

        for metrics_set in metrics:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {group_name}:{group.metrics_group_id}')
                
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self._client.invoke,
                        FunctionName=self.metrics_lambda_name,
                        Payload=json.dumps({
                            'group': group_name,
                            **group.record,
                            **group.custom
                        })
                    )
                )
            
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Lambda - {self.metrics_lambda_name}')


    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Lambda - {self.metrics_lambda_name}')

        for metrics_set in metrics_sets:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.custom_metrics.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {group_name}:{metrics_set.metrics_set_id}')

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self._client.invoke,
                        FunctionName=self.metrics_lambda_name,
                        Payload=json.dumps({
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': group_name,
                            **group
                        })
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Lambda - {self.metrics_lambda_name}')

    async def submit_errors(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics to Lambda - {self.metrics_lambda_name}')

        for metrics_set in metrics:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Errors Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self._client.invoke,
                        FunctionName=self.metrics_lambda_name,
                        Payload=json.dumps({
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            **error
                        })
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics to Lambda - {metrics_set.name}:{self.metrics_lambda_name}')
            

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')


from typing import Dict, Union
from .async_filesystem_logger import AsyncFilesystemLogger
from .sync_filesystem_logger import SyncFilesystemLogger    
from .async_logger import AsyncLogger
from .sync_logger import SyncLogger
from .async_spinner import AsyncSpinner
from .logger_types import LoggerTypes


class LoggerTypesMap:
    
    def __init__(self) -> None:
        self.logger_types={}

        for logger_type in LoggerTypes:
            logger_name = logger_type.name.lower()
            self.logger_types[logger_name] = logger_type

        self.async_loggers: Dict[LoggerTypes, Union[AsyncLogger, AsyncFilesystemLogger]] = {
            LoggerTypes.CONSOLE: AsyncLogger,
            LoggerTypes.DISTRIBUTED: AsyncLogger,
            LoggerTypes.HEDRA: AsyncLogger,
            LoggerTypes.FILESYSTEM: AsyncFilesystemLogger,
            LoggerTypes.DISTRIBUTED_FILESYSTEM: AsyncFilesystemLogger,
            LoggerTypes.SPINNER: AsyncSpinner
        }

        self.sync_loggers: Dict[LoggerTypes, Union[SyncLogger, SyncFilesystemLogger]] = {
            LoggerTypes.CONSOLE: SyncLogger,
            LoggerTypes.DISTRIBUTED: SyncLogger,
            LoggerTypes.HEDRA: SyncLogger,
            LoggerTypes.FILESYSTEM: SyncFilesystemLogger,
            LoggerTypes.DISTRIBUTED_FILESYSTEM: SyncFilesystemLogger
        }

        self.logger_names = {
            logger_type: logger_name for logger_name, logger_type in self.logger_types.items()
        }

    @property
    def names(self):
        return list(self.logger_types.keys())

    @property
    def types(self):
        return list(self.logger_types.values())

    def get_name(self, logger_type: LoggerTypes):
        return self.logger_names.get(logger_type)

    def get_type(self, logger_name: str):
        return self.logger_types.get(logger_name)


from __future__ import annotations
import sys
import asyncio
import time
import signal
import functools
import inspect
from enum import Enum
from os import get_terminal_size
from typing import List, Mapping, Any, Dict, Coroutine
from asyncio import Task
from yaspin.core import Yaspin
from yaspin.spinners import Spinners
from aiologger.levels import LogLevel
from aiologger.formatters.base import Formatter
from hedra.logging.spinner import ProgressText
from yaspin.helpers import to_unicode
from .async_logger import AsyncLogger
from .logger_types import LoggerTypes


class LoggerMode(Enum):
    CONSOLE='console'
    SYSTEM='system'


async def default_handler(signame: str,spinner: AsyncSpinner):  # pylint: disable=unused-argument
    """Signal handler, used to gracefully shut down the ``spinner`` instance
    when specified signal is received by the process running the ``spinner``.

    ``signum`` and ``frame`` are mandatory arguments. Check ``signal.signal``
    function for more details.
    """
    await spinner.fail()
    await spinner.stop()


class AsyncSpinner(Yaspin):

    def __init__(
        self, 
        logger_name: str=None,
        logger_type: LoggerTypes=LoggerTypes.SPINNER,
        log_level: LogLevel=LogLevel.NOTSET,
        logger_enabled: bool=True,
        spinner: Spinners=None, 
        text: ProgressText=None, 
        color: str=None, 
        on_color: str=None, 
        attrs: List[str]=None, 
        reversal: bool=False, 
        side: str="left", 
        sigmap: Dict[signal.Signals, Coroutine]=None, 
        timer: bool=False,
        enabled: bool=True
    ):
        super().__init__(
            spinner, 
            text, 
            color, 
            on_color, 
            attrs, 
            reversal, 
            side, 
            sigmap, 
            timer
        )

        self.logger: AsyncLogger = AsyncLogger(
            logger_name=logger_name,
            logger_type=logger_type,
            log_level=log_level,
            logger_enabled=logger_enabled
        )

        self.logger.initialize('%(message)s')
        self.display = text

        self.enabled = enabled
        self.logger_enabled = True
        self.logger_mode = LoggerMode.CONSOLE

        self._stdout_lock = asyncio.Lock()
        self._loop = asyncio.get_event_loop()

    @property
    def console(self):
        if self.logger_mode == LoggerMode.SYSTEM:
            for handler in self.logger.handlers:
                handler.formatter = Formatter('%(message)s')

            self.logger_mode = LoggerMode.CONSOLE

        return self

    @property
    def system(self):
        if self.logger_mode == LoggerMode.CONSOLE:
            for handler in self.logger.handlers:
                handler.formatter = Formatter(
                        '%(asctime)s - %(name)s - %(levelname)s - %(module)s:%(funcName)s:%(lineno)d - %(message)s',
                        datefmt='%Y-%m-%dT%H:%M:%S.%Z'
                )

            self.logger_mode = LoggerMode.SYSTEM

        return self

    def append_message(self, message: str) -> Coroutine[None]:
        return self.display.append_cli_message(message)

    def set_default_message(self, message: str) -> Coroutine[None]:
        return self.display.clear_and_replace(message)

    def set_message_at(self, message_index: int, message:str) -> None:
        if message_index < len(self.display.cli_messages):
            self.display.cli_messages[message_index] = message

    def finalize(self):
        self.display.finalized = True

    def group_finalize(self):
        self.display.group_finalized = True

    async def debug(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:
        
        if self.logger_enabled:
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.debug(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()

            return log_result

    async def info(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:

        if self.logger_enabled:
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.info(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()
            
            return log_result

    async def warning(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:
        
        if self.logger_enabled:
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.warning(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()
            return log_result

    async def warn(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:
        if self.logger_enabled:  
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.warn(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()
            return log_result

    async def error(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:
        
        if self.logger_enabled:
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.error(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()
            return log_result

    async def critical(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:
        
        if self.logger_enabled:
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.critical(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()
            return log_result
    
    async def fatal(self, message: str, *args: List[Any], **kwargs: Mapping[str, Any]) -> Task:
        
        if self.logger_enabled:
            await self._stdout_lock.acquire()
            await self._clear_line()

            # Ensure output is Unicode

            log_result = await self.logger.fatal(
                message,
                *args,
                **kwargs
            )

            self._cur_line_len = 0
            self._stdout_lock.release()
            return log_result

    async def __aenter__(self):
        self.display.group_timer.reset()
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, traceback):
        # Avoid stop() execution for the 2nd time
        if self.enabled and self._spin_thread.done() is False and self._spin_thread.cancelled() is False:
            await self.stop()
            
        return False  # nothing is handled

    def __call__(self, fn):
        @functools.wraps(fn)
        async def inner(*args, **kwargs):
            async with self:

                if inspect.iscoroutinefunction(fn):
                    return await fn(*args, **kwargs)

                else:
                    return fn(*args, **kwargs)

        return inner

    async def start(self):
        if self.enabled:

            if self._sigmap:
                self._register_signal_handlers()
            
            await self._hide_cursor()
            self._start_time = time.time()
            self._stop_time = None  # Reset value to properly calculate subsequent spinner starts (if any)  # pylint: disable=line-too-long
            self._stop_spin = asyncio.Event()
            self._hide_spin = asyncio.Event()
            try:
                self._spin_thread = asyncio.create_task(self._spin())
            finally:
                # Ensure cursor is not hidden if any failure occurs that prevents
                # getting it back
                await self._show_cursor()

            self.display.start_cli_tasks()

    async def stop(self):
        if self.enabled:
            self._stop_time = time.time()

            if self._dfl_sigmap:
                # Reset registered signal handlers to default ones
                self._reset_signal_handlers()

            if self._spin_thread:
                self._stop_spin.set()
                await self._spin_thread

            await self._clear_line()
            await self.display.stop_cli_tasks()
            await self._show_cursor()

    async def hide(self):
        """Hide the spinner to allow for custom writing to the terminal."""
        thr_is_alive = self._spin_thread and (self._spin_thread.done() is False and self._spin_thread.cancelled() is False)

        if thr_is_alive and not self._hide_spin.is_set():
            
            # set the hidden spinner flag
            self._hide_spin.set()
            await self._clear_line()

            # flush the stdout buffer so the current line
            # can be rewritten to
            await self._loop.run_in_executor(
                None,
                sys.stdout.flush
            )  

    async def show(self):
        """Show the hidden spinner."""
        thr_is_alive = self._spin_thread and (self._spin_thread.done() is False and self._spin_thread.cancelled() is False)

        if thr_is_alive and self._hide_spin.is_set():
            
            # clear the hidden spinner flag
            self._hide_spin.clear()

            # clear the current line so the spinner is not appended to it
            await self._clear_line()      

    async def write(self, text):
        if self.logger_enabled:
            """Write text in the terminal without breaking the spinner."""
            # similar to tqdm.write()
            # https://pypi.python.org/pypi/tqdm#writing-messages
            await self._stdout_lock.acquire()
            await self._clear_line()

            if isinstance(text, (str, bytes)):
                _text = to_unicode(text)
            else:
                _text = str(text)

            # Ensure output is Unicode
            assert isinstance(_text, str)

            await self._loop.run_in_executor(
                None,
                sys.stdout.write,
                
            )

            self._cur_line_len = 0    
            self._stdout_lock.release()  

    async def ok(self, text="OK"):
        if self.enabled:
            await self.display.stop_cli_tasks()
            """Set Ok (success) finalizer to a spinner."""
            _text = text if text else "OK"
            await self._freeze(_text)

    async def fail(self, text="FAIL"):
        if self.enabled:
            await self.display.stop_cli_tasks()
            """Set fail finalizer to a spinner."""
            _text = text if text else "FAIL"
            await self._freeze(_text)

    async def _freeze(self, final_text):
        """Stop spinner, compose last frame and 'freeze' it."""
        text = to_unicode(final_text)
        self._last_frame = self._compose_out(text, mode="last")

        # Should be stopped here, otherwise prints after
        # self._freeze call will mess up the spinner
        await self.stop()
        

        await self._loop.run_in_executor(
            None,
            sys.stdout.write,
            self._last_frame
        )

        self._cur_line_len = 0  

    async def _spin(self):
        while not self._stop_spin.is_set():

            if self._hide_spin.is_set():
                # Wait a bit to avoid wasting cycles
                await asyncio.sleep(self._interval)
                continue
            
            await self._stdout_lock.acquire()
            terminal_size = await self._loop.run_in_executor(
                None,
                get_terminal_size
            )

            terminal_width = terminal_size[0]

            # Compose output
            spin_phase = next(self._cycle)
            out = self._compose_out(spin_phase)

            if len(out) > terminal_width:
                out = f'{out[:terminal_width-1]}...'

            # Write
               
            await self._clear_line()
            
            await self._loop.run_in_executor(
                None,
                sys.stdout.write,
                out
            )

            await self._loop.run_in_executor(
                None,
                sys.stdout.flush
            )

            self._cur_line_len = max(self._cur_line_len, len(out))

            # Wait
            try:
                await asyncio.wait_for(self._stop_spin.wait(), timeout=self._interval)

            except asyncio.TimeoutError:
                pass

            self._stdout_lock.release()

    async def _clear_line(self):
        if sys.stdout.isatty():
            # ANSI Control Sequence EL does not work in Jupyter
            await self._loop.run_in_executor(
                None,
                sys.stdout.write,
                "\r\033[K"
            )

        else:
            fill = " " * self._cur_line_len
            await self._loop.run_in_executor(
                None,
                sys.stdout.write,
                sys.stdout.write,
                f"\r{fill}\r"
            )
            
    @staticmethod
    async def _show_cursor():
        loop = asyncio.get_event_loop()
        if sys.stdout.isatty():
            # ANSI Control Sequence DECTCEM 1 does not work in Jupyter
            await loop.run_in_executor(
                None,
                sys.stdout.write,
                "\033[?25h"
            )

            await loop.run_in_executor(
                None,
                sys.stdout.flush
            )

    @staticmethod
    async def _hide_cursor():
        loop = asyncio.get_event_loop()
        if sys.stdout.isatty():
            # ANSI Control Sequence DECTCEM 1 does not work in Jupyter
            await loop.run_in_executor(
                None,
                sys.stdout.write,
                "\033[?25l"
            )

            await loop.run_in_executor(
                None,
                sys.stdout.flush
            )

    def _register_signal_handlers(self):
        # SIGKILL cannot be caught or ignored, and the receiving
        # process cannot perform any clean-up upon receiving this
        # signal.
        if signal.SIGKILL in self._sigmap:
            raise ValueError(
                "Trying to set handler for SIGKILL signal. "
                "SIGKILL cannot be caught or ignored in POSIX systems."
            )

        for sig, sig_handler in self._sigmap.items():
            # A handler for a particular signal, once set, remains
            # installed until it is explicitly reset. Store default
            # signal handlers for subsequent reset at cleanup phase.
            dfl_handler = signal.getsignal(sig)
            self._dfl_sigmap[sig] = dfl_handler

            # ``signal.SIG_DFL`` and ``signal.SIG_IGN`` are also valid
            # signal handlers and are not callables.
            if callable(sig_handler):
                # ``signal.signal`` accepts handler function which is
                # called with two arguments: signal number and the
                # interrupted stack frame. ``functools.partial`` solves
                # the problem of passing spinner instance into the handler
                # function.
                sig_handler = functools.partial(sig_handler, spinner=self)

            self._loop.add_signal_handler(getattr(signal, sig.name),
                            lambda signame=sig.name: asyncio.create_task(sig_handler(self)))

    def _reset_signal_handlers(self):
        for sig, sig_handler in self._dfl_sigmap.items():
            self._loop.add_signal_handler(getattr(signal, sig.name),
                            lambda signame=sig.name: asyncio.create_task(sig_handler(signame, self)))




from asyncio import (
    Future,
    Protocol,
    get_event_loop
)


class FlowControlMixin(Protocol):
    """Reusable flow control logic for StreamWriter.drain().
    This implements the protocol methods pause_writing(),
    resume_writing() and connection_lost().  If the subclass overrides
    these it must call the super methods.
    StreamWriter.drain() must wait for _drain_helper() coroutine.
    """
    __slots__ = (
        '_loop',
        '_paused',
        '_drain_waiter',
        '_connection_lost'
    )

    def __init__(self, loop=None):

        if loop:
            self._loop = loop
        else:
            self._loop = get_event_loop()

        self._paused = False
        self._drain_waiter: Future = None
        self._connection_lost = False

    def pause_writing(self):
        assert not self._paused
        self._paused = True

    def resume_writing(self):
        assert self._paused
        self._paused = False

        waiter = self._drain_waiter
        if waiter is not None:
            self._drain_waiter = None
            if not waiter.done():
                waiter.set_result(None)

    def connection_lost(self, exc):
        self._connection_lost = True
        # Wake up the writer if currently paused.
        if not self._paused:
            return
        waiter = self._drain_waiter
        if waiter is None:
            return
        self._drain_waiter = None
        if waiter.done():
            return
        if exc is None:
            waiter.set_result(None)
        else:
            waiter.set_exception(exc)

    async def _drain_helper(self):
        if self._connection_lost:
            raise ConnectionResetError('Connection lost')
        if not self._paused:
            return
        waiter = self._drain_waiter
        assert waiter is None or waiter.cancelled()
        waiter = self._loop.create_future()
        self._drain_waiter = waiter
        await waiter

    def _get_close_waiter(self, stream):
        raise NotImplementedError


from hedra.core.engines.types.playwright import PlaywrightResult
from .base_event import BaseEvent


class PlaywrightEvent(BaseEvent):

    __slots__ = (
        'event_id',
        'action_id',
        'url',
        'headers',
        'command',
        'selector',
        'x_coord',
        'y_coord',
        'frame',
        'timings'
    )

    def __init__(self, result: PlaywrightResult) -> None:
        super(
            PlaywrightEvent,
            self
        ).__init__(result)

        self.url = result.url
        self.headers = result.headers
        self.command = result.command
        self.selector = result.selector
        self.x_coord = result.x_coord
        self.y_coord = result.y_coord
        self.frame = result.frame

        self.time = result.complete - result.start

        self.timings = {
            'total': self.time,
            'waiting': result.start - result.wait_start,
            'connecting': result.connect_end - result.start,
            'writing': result.write_end - result.connect_end,
            'reading': result.complete - result.write_end
        }


import asyncio
import functools
import datetime
import json
import uuid
from typing import List

import psutil
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet
from concurrent.futures import ThreadPoolExecutor

try:
    import boto3
    from .cloudwatch_config import CloudwatchConfig
    has_connector = True

except Exception:
    boto3 = None
    CloudwatchConfig = None
    has_connector = False


class Cloudwatch:

    def __init__(self, config: CloudwatchConfig) -> None:
        self.aws_access_key_id = config.aws_access_key_id
        self.aws_secret_access_key = config.aws_secret_access_key
        self.region_name = config.region_name
        self.iam_role_arn = config.iam_role_arn
        self.schedule_rate = config.schedule_rate
        self.cloudwatch_targets = config.cloudwatch_targets
        self.aws_resource_arns = config.aws_resource_arns
        self.submit_timeout = config.submit_timeout
        self.events_rule_name = config.events_rule
        self.metrics_rule_name = config.metrics_rule
        self.errors_rule_name = 'stage_errors'

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self.events_rule = None
        self.metrics_rule= None

        self.client = None
        self._loop = asyncio.get_event_loop()

    async def connect(self):
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating AWS Cloudwatch client for region - {self.region_name}')

        self.client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'events',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created AWS Cloudwatch client for region - {self.region_name}')

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Cloudwatch rule - {self.events_rule_name}')

        cloudwatch_events = [
            {
                'Time': datetime.datetime.now(),
                'Detail': json.dumps(event.record),
                'DetailType': self.events_rule_name,
                'Resources': self.aws_resource_arns,
                'Source': self.events_rule_name
            } for event in events
        ]

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=cloudwatch_events
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Cloudwatch rule - {self.events_rule_name}')

    
    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Cloudwatch rule - {self.metrics_rule_name}')

        common_metrics = [
            {
                'Time': datetime.datetime.now(),
                'Detail': json.dumps({
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'common',
                    **metrics_set.common_stats
                }),
                'DetailType': self.events_rule_name,
                'Resources': self.aws_resource_arns,
                'Source': self.metrics_rule_name
            } for metrics_set in metrics_sets
        ]

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=common_metrics
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Cloudwatch rule - {self.metrics_rule_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Cloudwatch rule - {self.metrics_rule_name}')

        metrics = []
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {group_name}:{group.metrics_group_id}')
                
                metrics.append({
                    'Time': datetime.datetime.now(),
                    'Detail': json.dumps({
                        **group.record,
                        'group': group_name
                    }),
                    'DetailType': self.metrics_rule_name,
                    'Resources': self.aws_resource_arns,
                    'Source': self.metrics_rule_name
                })

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=metrics
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Cloudwatch rule - {self.metrics_rule_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Cloudwatch rule - {self.metrics_rule_name}')

        custom_metrics = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_group_name, group in metrics_set.custom_metrics.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {custom_group_name}')

                custom_metrics.append({
                    'Time': datetime.datetime.now(),
                    'Detail': json.dumps({
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': custom_group_name,
                        **group
                    }),
                    'DetailType': self.metrics_rule_name,
                    'Resources': self.aws_resource_arns,
                    'Source': self.metrics_rule_name
                })

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=custom_metrics
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Cloudwatch rule - {self.metrics_rule_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Cloudwatch rule - {self.errors_rule_name}')

        cloudwatch_errors = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                cloudwatch_errors.append({
                    'Time': datetime.datetime.now(),
                    'Detail': json.dumps(error),
                    'DetailType': self.errors_rule_name,
                    'Resources': self.aws_resource_arns,
                    'Source': self.errors_rule_name
                })

        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=cloudwatch_errors
                )
            ),
            timeout=self.submit_timeout
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Cloudwatch rule - {self.errors_rule_name}')


    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')



import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    PrometheusConfig
)


class SubmitPrometheusResultsStage(Submit):
    config=PrometheusConfig(
        pushgateway_address='localhost:9091',
        auth_request_method='GET',
        username=os.getenv('PROMETHEUS_PUSHGATEWAY_USERNAME', ''),
        password=os.getenv('PROMETHEUS_PUSHGATEWAY_PASSWORD', ''),
        namespace='results',
        job_name='results'
    )


import binascii
from typing import Dict, Iterator, List, Union
from hedra.core.engines.types.common.hooks import Hooks
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.http2.action import HTTP2Action


class GRPCAction(HTTP2Action):

    def __init__(
        self,
        name: str, 
        url: str, 
        method: str = 'GET', 
        headers: Dict[str, str] = {}, 
        data: Union[str, dict, Iterator, bytes, None] = None, 
        user: str=None, 
        tags: List[Dict[str, str]] = []
    ) -> None:
    
        super(
            GRPCAction, 
            self
        ).__init__(
            name, 
            url, 
            method, 
            headers, 
            data, user, 
            tags
        )

        self.timeout = 60
        self.type = RequestTypes.GRPC
        self.hooks: Hooks[GRPCAction] = Hooks()

    def _setup_headers(self):
        grpc_headers = {
            'Content-Type': 'application/grpc',
            'Grpc-Timeout': f'{self.timeout}S',
            'TE': 'trailers'
        }
        self._headers.update(grpc_headers)

        super(
            GRPCAction,
            self
        )._setup_headers()

    def _setup_data(self) -> None:
        encoded_protobuf = str(binascii.b2a_hex(self.data.SerializeToString()), encoding='raw_unicode_escape')
        encoded_message_length = hex(int(len(encoded_protobuf)/2)).lstrip("0x").zfill(8)
        encoded_protobuf = f'00{encoded_message_length}{encoded_protobuf}'

        self.encoded_data = binascii.a2b_hex(encoded_protobuf)


import time
import uuid
import asyncio
from hedra.core.engines.types.http import MercuryHTTPClient
from hedra.core.engines.types.http.connection import HTTPConnection
from hedra.core.engines.types.common import Timeouts
from typing import Awaitable, Union
from .action import GraphQLAction
from .result import GraphQLResult


GraphQLResponseFuture = Awaitable[Union[GraphQLAction, Exception]]


class MercuryGraphQLClient(MercuryHTTPClient):

    def __init__(
        self, 
        concurrency: int = 10 ** 3, 
        timeouts: Timeouts = Timeouts(), 
        reset_connections: bool = False
    ) -> None:

        super(
            MercuryGraphQLClient,
            self
        ).__init__(
            concurrency, 
            timeouts, 
            reset_connections
        )

        self.session_id = str(uuid.uuid4())

    async def execute_prepared_request(self, action: GraphQLAction) -> GraphQLResponseFuture:
   
        response = GraphQLResult(action)
        response.wait_start = time.monotonic()
        self.active += 1

        async with self.sem:
            connection = self.pool.connections.pop()
            
            try:

                if action.hooks.listen:
                    event = asyncio.Event()
                    action.hooks.channel_events.append(event)
                    await event.wait()

                if action.hooks.before:
                    action = await action.hooks.before(action, response)
                    action.setup()

                response.start = time.monotonic()

                await connection.make_connection(
                    action.url.hostname,
                    action.url.ip_addr,
                    action.url.port,
                    action.url.socket_config,
                    timeout=self.timeouts.connect_timeout,
                    ssl=action.ssl_context
                )

                response.connect_end = time.monotonic()

                connection.write(action.encoded_headers)
                
                if action.encoded_data:
                    if action.is_stream:
                        action.write_chunks(connection)

                    else:
                        connection.write(action.encoded_data)

                response.write_end = time.monotonic()

                response.response_code = await connection.reader.readline_fast()
                headers = await connection.read_headers()
                status = response.status

                if status >= 300 and status < 400:
                    for _ in range(action.redirects):

                        redirect_url = headers.get(b'Location')

                        redirect_action = GraphQLAction(
                            action.name,
                            redirect_url,
                            method=action.method
                        )

                        redirect_action.is_setup = True

                        self.prepare(redirect_action)

                        redirect_action.encoded_headers = action.encoded_headers
                        redirect_action.encoded_data = action.encoded_data

                        await connection.make_connection(
                            redirect_action.url.hostname,
                            redirect_action.url.ip_addr,
                            redirect_action.url.port,
                            redirect_action.url.socket_config,
                            timeout=self.timeouts.connect_timeout,
                            ssl=redirect_action.ssl_context
                        )

                        response.connect_end = time.monotonic()

                        connection.write(redirect_action.encoded_headers)
                        
                        if redirect_action.encoded_data:
                            if redirect_action.is_stream:
                                redirect_action.write_chunks(connection)

                            else:
                                connection.write(redirect_action.encoded_data)

                        response.write_end = time.monotonic()

                        response.response_code = await connection.reader.readline_fast()
                        headers = await connection.read_headers()

                        status = response.status

                        if status >= 200 and status < 300:
                            break

                content_length = headers.get(b'content-length')
                transfer_encoding = headers.get(b'transfer-encoding')
    
                # We require Content-Length or Transfer-Encoding headers to read a
                # request body, otherwise it's anyone's guess as to how big the body
                # is, and we ain't playing that game.
                body = bytearray()
                if content_length:
                    body = await connection.readexactly(int(content_length))

                elif transfer_encoding:
                    
                    all_chunks_read = False

                    while True and not all_chunks_read:

                        chunk_size = int((await connection.readuntil()).rstrip(), 16)
                        if not chunk_size:
                            # read last CRLF
                            body.extend(
                                await connection.readuntil()
                            )
                            break
                        
                        chunk = await connection.readexactly(chunk_size + 2)
                        body.extend(
                            chunk[:-2]
                        )

                    all_chunks_read = True

                response.complete = time.monotonic()
                response.headers = headers
                response.body = body
                
                self.pool.connections.append(connection)

                if action.hooks.after:
                    action = await action.hooks.after(action, response)
                    action.setup()

                if action.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel(response, action.hooks.listeners)
                        ) for channel in action.hooks.channels
                    ])

                    for listener in action.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            listener.setup()
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()    

            except Exception as e:
                response.complete = time.monotonic()
                response.error = str(e)
                
                self.pool.connections.append(HTTPConnection(reset_connection=self.pool.reset_connections))

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            return response


from typing import Dict
from hedra.reporting.types.common.types import ReporterTypes
from pydantic import BaseModel


class DogStatsDConfig(BaseModel):
    host: str='localhost'
    port: int=8125
    custom_fields: Dict[str, str]={}
    reporter_type: ReporterTypes=ReporterTypes.DogStatsD


import os
import struct
from typing import Tuple
from base64 import encodebytes as base64encode
from .connection import WebsocketConnection
from .constants import HEADER_LENGTH_INDEX

def create_sec_websocket_key():
    randomness = os.urandom(16)
    return base64encode(randomness).decode('utf-8').strip()

def pack_hostname(hostname):
    # IPv6 address
    if ':' in hostname:
        return '[' + hostname + ']'

    return hostname


def get_header_bits(raw_headers: bytes):

    b1 = raw_headers[0]
    fin = b1 >> 7 & 1
    rsv1 = b1 >> 6 & 1
    rsv2 = b1 >> 5 & 1
    rsv3 = b1 >> 4 & 1
    opcode = b1 & 0xf
    b2 = raw_headers[1]
    has_mask = b2 >> 7 & 1
    length_bits = b2 & 0x7f

    header_bits = (fin, rsv1, rsv2, rsv3, opcode, has_mask, length_bits)

    return header_bits


async def get_message_buffer_size(header_bits: Tuple[int], connection: WebsocketConnection):
        bits = header_bits[HEADER_LENGTH_INDEX]
        length_bits = bits & 0x7f
        length = 0
        if length_bits == 0x7e:
            v = await connection.readexactly(2)
            length = struct.unpack("!H", v)[0]
        elif length_bits == 0x7f:
            v = await connection.readexactly(8)
            length = struct.unpack("!Q", v)[0]
        else:
            length = length_bits

        return length



from typing import Any, Dict, List
from hedra.plugins.types.engine import (
    EnginePlugin,
    Action,
    Result,
    connect,
    execute,
    close
)


class CustomAction(Action):

    def __init__(
        self, 
        name: str, 
        source: str = None, 
        user: str = None, 
        tags: List[Dict[str, str]] = [], 
        use_security_context: bool = False
    ) -> None:
        super(
            CustomAction,
            self
        ).__init__(
            name=name, 
            source=source, 
            user=user, 
            tags=tags, 
            use_security_context=use_security_context
        )


class CustomResult(Result):

    def __init__(
        self, 
        action: CustomAction, 
        error: Exception = None
    ) -> None:
        super().__init__(
            action, 
            error
        )


class CustomEngine(EnginePlugin[CustomAction, CustomResult]):
    cache: Dict[str, Any] = {}
    action = CustomAction
    result = CustomResult

    @connect()
    async def connect_engine(self, action: CustomAction):
        pass

    @execute()
    async def execute_action(self, action: CustomAction, result: CustomResult):
        pass

    @close()
    async def close_engine(self):
        pass


from .async_filesystem_logger import AsyncFilesystemLogger
from .async_logger import AsyncLogger
from .async_spinner import AsyncSpinner
from .logger import Logger
from .logger_types import LoggerTypes
from .logger_types_map import LoggerTypesMap
from .sync_filesystem_logger import SyncFilesystemLogger
from .sync_logger import SyncLogger


from .proto import (
    DistributedServerServicer,
    PipelineStageResponse,
    WorkerUpdateResponse,
    WorkerRegisterResponse,
    LeaderRegisterResponse,
    PollLeaderResponse,
    NewJobLeaderResponse,
    LeaderHeartbeatResponse,
    WorkerDeregisterResponse,
    LeaderJobResponse
)
from google.protobuf.json_format import (
    MessageToDict
)

from .worker_registry import NewWorkerQueue
from .job_registry import JobRegistry


class LeaderService(DistributedServerServicer):

    def __init__(self):
        super().__init__()
        
        self.leader_ip = None
        self.leader_port = None
        self.leader_address = None
        self.running = False
        self.new_worker_queue = NewWorkerQueue()
        self.job_registry = JobRegistry()

    async def setup_leader_addresses(self, config):
        await self.job_registry.setup_leader_addresses(config)

        self.leader_ip = self.job_registry.leader_ip
        self.leader_port = self.job_registry.leader_port
        self.leader_address = self.job_registry.leader_address  

    async def setup(self, config, leader_id):
        self.leader_id = leader_id

        await self.job_registry.setup(
            config,
            leader_id
        )      

    async def MarkStageCompleted(self, stage_completed_request, context):

        job = self.job_registry[stage_completed_request.job_id]

        if job:
            await job.pipeline.update_completed(stage_completed_request)

        return PipelineStageResponse(
            host_address=self.job_registry.leader_address,
            host_id=str(self.job_registry.leader_id)
        )

    async def CreateNewJob(self, new_job, context):
        
        config = MessageToDict(new_job.job_config)
        leader_job_id = config.get('job_name')

        if self.job_registry.ready:
            if leader_job_id:
                await self.job_registry.create_job(
                    leader_job_id, 
                    config, 
                    timeout=new_job.job_timeout
                )
                
                await self.job_registry.start_job(leader_job_id)

        return NewJobLeaderResponse(
            host_address=self.job_registry.leader_address,
            host_id=str(self.job_registry.leader_id),
            job_id=leader_job_id
        )

    async def AddClient(self, new_worker_request, context):
        await self.job_registry.workers.registry.register_worker(new_worker_request.server_address)
        
        response = WorkerRegisterResponse(
            host_address=self.job_registry.leader_address,
            host_id=str(self.job_registry.leader_id)
        )

        response.job_config.update({'batch_size': 0, 'batch_time': 0, 'total_time': '00:00:00'})
            
        return response

    async def GetUpdate(self, update_request, context):
        update = await self.job_registry.workers.get_actions_completed()

        response = WorkerUpdateResponse(
            leader_address=self.job_registry.leader_address,
            leader_id=str(self.job_registry.leader_id)
        )
        
        response.completed.update(update)
        return response

    async def RegisterLeader(self, leader_registration_request, context):
        host_config = await self.job_registry.electorate.register(leader_registration_request)     
        return LeaderRegisterResponse(**host_config)

    async def PollLeader(self, request, context):
        return PollLeaderResponse(
            host_address=self.job_registry.leader_address,
            host_id=str(self.job_registry.leader_id),
            completed=self.job_registry.electorate.poll_completed
        )

    async def PollLeaderJob(self, poll_leader_job, context):
        job = self.job_registry[poll_leader_job.job_id]

        completed = False
        if job:
            completed = job.pipeline.current_stage_completed

        return PollLeaderResponse(
            host_address=self.job_registry.leader_address,
            host_id=str(self.job_registry.leader_id),
            completed=completed
        )

    async def CheckLeaderHeartbeat(self, request, context):
        return LeaderHeartbeatResponse(
            host_address=self.job_registry.leader_address,
            host_id=str(self.job_registry.leader_id),
            status='OK'
        )

    async def RemoveClient(self, deregister_request, context):

        self.session_logger.info(f'\nClient from - {deregister_request.host_address} - has requested disconnect.')
        await self.job_registry.stop_job_monitor()

        self.session_logger.debug('Removing worker...')
        await self.job_registry.workers.registry.remove_worker(deregister_request.host_address)

        response = WorkerDeregisterResponse(
            host_address=self.job_registry.leader_address,
            host_id=str(self.job_registry.leader_id)
        )

        self.session_logger.info('Client de-registered.')
        await self.job_registry.start_job_monitor()

        return response

    async def ShareCurrentJob(self, shared_job_request, context):
        job_config = MessageToDict(shared_job_request.job_config)
        job_pipeline_status = MessageToDict(shared_job_request.pipeline_status)

        current_job_config = self.job_registry[shared_job_request.job_id]
        if current_job_config is None:
            await self.job_registry.create_job(shared_job_request.job_id, job_config)
            await self.job_registry.start_job(shared_job_request.job_id)
            updated_pipeline_status = await self.job_registry[shared_job_request.job_id].pipeline.update_pipeline_status(job_pipeline_status)

        else:
            updated_pipeline_status = await self.job_registry[shared_job_request.job_id].pipeline.update_pipeline_status(job_pipeline_status)

        host_job = self.job_registry[shared_job_request.job_id]

        response = LeaderJobResponse(
            host_address=self.job_registry.leader_address,
            host_id=str(self.job_registry.leader_id),
            job_id=shared_job_request.job_id
        )

        response.job_config.update(host_job.config)
        response.pipeline_status.update(updated_pipeline_status)

        return response


import time
import uuid
import asyncio
import psutil
from hedra.core.personas.types.default_persona import DefaultPersona
from hedra.core.engines.client.config import Config
from hedra.core.personas.types.types import PersonaTypes


class RampedIntervalPersona(DefaultPersona):

    def __init__(self, config: Config):
        super(RampedIntervalPersona, self).__init__(config)

        self.persona_id = str(uuid.uuid4())
        self.type = PersonaTypes.RAMPED_INTERVAL
            
    async def generator(self, total_time):
        elapsed = 0
        idx = 0
        max_pool_size = int(self.batch.size * (psutil.cpu_count(logical=False) * 2)/self.workers)
        generation_batch_interval = self.batch.interval * self.batch.gradient

        start = time.time()
        batch_start = time.time()
        while elapsed < total_time:
            action_idx = idx%self.actions_count
            yield action_idx
            
            await asyncio.sleep(0)
            elapsed = time.time() - start
            batch_elapsed = time.time() - batch_start
            idx += 1

            if batch_elapsed >= generation_batch_interval:
                increase_amount = (self.batch.interval * self.batch.gradient)
                next_batch_time = generation_batch_interval + increase_amount

                if next_batch_time < self.batch.interval:
                    generation_batch_interval = next_batch_time
                else:
                    generation_batch_interval = self.batch.interval
                
                await asyncio.sleep(self.batch.interval)
                batch_start = time.time()

            if self._hooks[action_idx].session.active%max_pool_size == 0:
                    try:
                        max_wait = total_time - elapsed
                        await asyncio.wait_for(
                            self._hooks[action_idx].session.wait_for_active_threshold(),
                            timeout=max_wait
                        )
                    except asyncio.TimeoutError:
                        pass



import asyncio

from hedra.runners.leader_services.job_registry import job

from hedra.runners.worker_services.proto.worker_pb2 import JobCompleteResponse
from .leader_registry import NewLeaderQueue
from .proto import (
    WorkerServerServicer,
    WorkerServerUpdateResponse,
    NewJobResponse,
    PollWorkerResponse,
    WorkerHeartbeatResponse,
    LeaderRegisterWorkerResponse
)
from google.protobuf.json_format import (
    MessageToDict
)
from .jobs_registry import JobRegistry


class WorkerService(WorkerServerServicer):

    def __init__(self, config, repoter_config, worker_address, worker_id):
        super().__init__()
        
        logger = Logger()
        self.session_logger = logger.generate_logger()
        self.worker_address = worker_address
        self.worker_id = worker_id

        self.new_leader_queue = NewLeaderQueue()
        self.job_registry = JobRegistry(
            config, 
            repoter_config, 
            self.worker_address, 
            self.worker_id
        )

    async def CreateNewWorkerJob(self, new_job_request, context):
        new_job = MessageToDict(new_job_request).get('jobConfig')
        job_id = new_job_request.job_id

        await self.job_registry.create_job(
            job_id,
            new_job,
            new_job_request.host_address,
            timeout=new_job_request.job_timeout
        )

        await self.job_registry.start_job(job_id)

        response = NewJobResponse(
            host_address=self.worker_address,
            host_id=str(self.worker_id),
            job_id=job_id
        )

        return response

    async def GetRequestsCompleted(self, update_request, context):

        response = WorkerServerUpdateResponse(
            host_address=self.worker_address,
            host_id=str(self.worker_id)
        )

        completed = {}
        async for job in self.job_registry.iter_active():
            if job.pipeline.status == 'active':
                actions_completed, time_elapsed = await job.pipeline.executor.get_completed()

                if time_elapsed < 1:
                    current_rps = 0

                else:
                    current_rps = actions_completed/time_elapsed

                completed[job.job_id] = {
                    'actions_completed': actions_completed,
                    'time_elapsed': time_elapsed,
                    'current_rps': current_rps
                }

        response.completed.update(completed)
        return response

    async def GetJobStatus(self, poll_worker_reqest, context):
        return PollWorkerResponse(
            host_address=self.worker_address,
            host_id=self.worker_id,
            completed=self._active_jobs[poll_worker_reqest.job_id]['task'].done()
        )

    async def GetJobResults(self, results_request, context):
        response = JobCompleteResponse(
            host_address=self.worker_address,
            host_id=str(self.worker_id)
        )

        job = await self.job_registry.get_from_active(results_request.job_id)

        if job and job.pipeline.status == 'completed':
            job_results = await job.get_task()
            response.job_results.update(job_results)
            await job.remove_leader(results_request.host_address)
            await self.job_registry.set_job_complete(job.job_id)

        return response

    async def CheckWorkerHeartbeat(self, request, context):
        return WorkerHeartbeatResponse(
            host_address=self.worker_address,
            host_id=str(self.worker_id),
            status='OK'
        )

    async def LeaderRegisterWorker(self, leader_register_request, context):
        await self.new_leader_queue.add(leader_register_request)

        async for job in self.job_registry.iter_active():
            await job.remove_leader(leader_register_request.host_address)

        return LeaderRegisterWorkerResponse(
            host_address=self.worker_address,
            host_id=str(self.worker_id),
            server_address=self.worker_address
        )




from typing import List, Union
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.registry.registry_types import (
    ActionHook,
    TaskHook,
    ChannelHook
)
from hedra.core.graphs.stages.validate.exceptions import HookValidationError
from .base_hook_validator import BaseHookVaidator


class ChannelHookValidator(BaseHookVaidator):

    def __init__(self, metadata_string: str) -> None:
        super(
            ChannelHookValidator,
            self
        ).__init__(metadata_string)

        self.action_and_task_hooks: List[Union[ActionHook, TaskHook]] = []

    async def validate(self, hook: ChannelHook):

        try:

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Validating {hook.hook_type.name.capitalize()} Hook - {hook.name}:{hook.hook_id}:{hook.hook_id}')

            assert hook.hook_type is HookType.CHANNEL, f"Hook type mismatch - hook {hook.name}:{hook.hook_id} is a {hook.hook_type.name} hook, but Hedra expected a {HookType.CHANNEL.name} hook."
            assert hook.shortname in hook.name, f"Shortname {hook.shortname} must be contained in full Hook name {hook.name}:{hook.hook_id} for @channel hook {hook.name}:{hook.hook_id}."
            assert hook.call is not None, f"Method is not not found on stage or was not supplied to @channel hook - {hook.name}:{hook.hook_id}"
            assert hook.call.__code__.co_argcount > 2, f"Missing required arguments 'result' or 'actions' for @channel hook {hook.name}:{hook.hook_id}"
            assert hook.call.__code__.co_argcount < 4, f"Too many args. - @channel hook {hook.name}:{hook.hook_id} only requires 'result' and 'actions' as additional args."
            assert 'self' in hook.call.__code__.co_varnames

            stage_actions: List[Union[ActionHook, TaskHook]] = list(
                filter(
                    lambda action_or_task: action_or_task.stage == hook.stage,
                    self.action_and_task_hooks
                )
            )

            stage_notifiers = []
            for action in stage_actions:
                stage_notifiers.extend(action.notifiers)

            stage_listeners = []
            for action in stage_actions:
                stage_listeners.extend(action.listeners)

            assert hook.shortname in stage_notifiers, f"No notifiers found. - @channel hook {hook.name}:{hook.hook_id} requires at least one action with a notify list containing its name"
            assert hook.shortname in stage_listeners, f"No listeners found. - @channel hook {hook.name}:{hook.hook_id} requires at least one action with a listeners list containing its name"

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Validated {hook.hook_type.name.capitalize()} Hook - {hook.name}:{hook.hook_id}:{hook.hook_id}')
        
        except AssertionError as hook_validation_error:
                raise HookValidationError(hook.stage_instance, str(hook_validation_error))


from ..stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_types import StageTypes
from .common import (
    idle_transition,
    invalid_transition,
    exit_transition,
    error_transition
)

from .validate import (
    validate_to_setup_transition,
    validate_to_wait_transition
)

from .idle import (
    invalid_idle_transition,
    idle_to_validate_transition,
    idle_to_wait_transition
)

from .setup import (
    setup_to_validate_transition,
    setup_to_optimize_transition,
    setup_to_execute_transition,
    setup_to_checkpoint_transition,
    setup_to_wait_transition
)

from .optimize import (
    optimize_to_execute_transition,
    optimize_to_checkpoint_transition,
    optimize_to_wait_transition
)

from .execute import (
    execute_to_setup_transition,
    execute_to_execute_transition,
    execute_to_optimize_transition,
    execute_to_teardown_transition,
    execute_to_analyze_transition,
    execute_to_checkpoint_transition,
    execute_to_wait_transition
)

from .teardown import (
    teardown_to_analyze_transition,
    teardown_to_checkpoint_transition,
    teardown_to_wait_transition
)

from .analyze import (
    analyze_to_checkpoint_transition,
    analyze_to_submit_transition,
    analyze_to_wait_transition
)

from .checkpoint import (
    checkpoint_to_setup_transition,
    checkpoint_to_optimize_transition,
    checkpoint_to_execute_transition,
    checkpoint_to_teardown_transition,
    checkpoint_to_analyze_transition,
    checkpoint_to_complete_transition,
    checkpoint_to_submit_transition,
    checkpoint_to_wait_transition,
    checkpoint_to_checkpoint_transition
)

from .submit import (
    submit_to_setup_transition,
    submit_to_optimize_transition,
    submit_to_execute_transition,
    submit_to_checkpoint_transition,
    submit_to_submit_transition,
    submit_to_complete_transition,
    submit_to_wait_transition
)


from .wait import (
    wait_to_wait_transition,
    wait_to_analyze_transition,
    wait_to_checkpoint_transition,
    wait_to_complete_transition,
    wait_to_error_transition,
    wait_to_execute_transition,
    wait_to_optimize_transition,
    wait_to_setup_transition,
    wait_to_submit_transition,
    wait_to_teardown_transition,
    wait_to_validate_transition
)


local_transitions = {

        # State: Idle
        (StageTypes.IDLE, StageTypes.IDLE): idle_transition,
        (StageTypes.IDLE, StageTypes.WAIT): idle_to_wait_transition,
        (StageTypes.IDLE, StageTypes.SETUP): invalid_idle_transition,
        (StageTypes.IDLE, StageTypes.VALIDATE): idle_to_validate_transition,
        (StageTypes.IDLE, StageTypes.OPTIMIZE): invalid_idle_transition,
        (StageTypes.IDLE, StageTypes.EXECUTE):  invalid_idle_transition,
        (StageTypes.IDLE, StageTypes.TEARDOWN): invalid_idle_transition,
        (StageTypes.IDLE, StageTypes.ANALYZE): invalid_idle_transition,
        (StageTypes.IDLE, StageTypes.CHECKPOINT): invalid_idle_transition,
        (StageTypes.IDLE, StageTypes.SUBMIT): invalid_idle_transition,
        (StageTypes.IDLE, StageTypes.COMPLETE): invalid_idle_transition,
        (StageTypes.IDLE, StageTypes.ERROR): error_transition,

        # State: Wait
        (StageTypes.WAIT, StageTypes.WAIT): wait_to_wait_transition,
        (StageTypes.WAIT, StageTypes.IDLE): invalid_transition,
        (StageTypes.WAIT, StageTypes.SETUP): wait_to_setup_transition,
        (StageTypes.WAIT, StageTypes.VALIDATE): wait_to_validate_transition,
        (StageTypes.WAIT, StageTypes.OPTIMIZE): wait_to_optimize_transition,
        (StageTypes.WAIT, StageTypes.EXECUTE): wait_to_execute_transition,
        (StageTypes.WAIT, StageTypes.TEARDOWN): wait_to_teardown_transition,
        (StageTypes.WAIT, StageTypes.ANALYZE): wait_to_analyze_transition,
        (StageTypes.WAIT, StageTypes.CHECKPOINT): wait_to_checkpoint_transition,
        (StageTypes.WAIT, StageTypes.SUBMIT): wait_to_submit_transition,
        (StageTypes.WAIT, StageTypes.COMPLETE): wait_to_complete_transition,
        (StageTypes.WAIT, StageTypes.ERROR): wait_to_error_transition,

        # State: Setup
        (StageTypes.SETUP, StageTypes.SETUP): invalid_transition,
        (StageTypes.SETUP, StageTypes.IDLE): invalid_transition,
        (StageTypes.SETUP, StageTypes.WAIT): setup_to_wait_transition,
        (StageTypes.SETUP, StageTypes.VALIDATE): setup_to_validate_transition,
        (StageTypes.SETUP, StageTypes.OPTIMIZE): setup_to_optimize_transition,
        (StageTypes.SETUP, StageTypes.EXECUTE): setup_to_execute_transition,
        (StageTypes.SETUP, StageTypes.TEARDOWN): invalid_transition,
        (StageTypes.SETUP, StageTypes.ANALYZE): invalid_transition,
        (StageTypes.SETUP, StageTypes.CHECKPOINT): setup_to_checkpoint_transition,
        (StageTypes.SETUP, StageTypes.SUBMIT): invalid_transition,
        (StageTypes.SETUP, StageTypes.COMPLETE): invalid_transition,
        (StageTypes.SETUP, StageTypes.ERROR): error_transition,

        # State: Validate
        (StageTypes.VALIDATE, StageTypes.VALIDATE): invalid_transition,
        (StageTypes.VALIDATE, StageTypes.IDLE): invalid_transition,
        (StageTypes.VALIDATE, StageTypes.WAIT): validate_to_wait_transition,
        (StageTypes.VALIDATE, StageTypes.SETUP): validate_to_setup_transition,
        (StageTypes.VALIDATE, StageTypes.OPTIMIZE): invalid_transition,
        (StageTypes.VALIDATE, StageTypes.EXECUTE): invalid_transition,
        (StageTypes.VALIDATE, StageTypes.TEARDOWN): invalid_transition,
        (StageTypes.VALIDATE, StageTypes.ANALYZE): invalid_transition,
        (StageTypes.VALIDATE, StageTypes.CHECKPOINT): invalid_transition,
        (StageTypes.VALIDATE, StageTypes.SUBMIT): invalid_transition,
        (StageTypes.VALIDATE, StageTypes.COMPLETE): invalid_transition,
        (StageTypes.VALIDATE, StageTypes.ERROR): error_transition,

        # State: Optimize
        (StageTypes.OPTIMIZE, StageTypes.OPTIMIZE): invalid_transition,
        (StageTypes.OPTIMIZE, StageTypes.IDLE): invalid_transition,
        (StageTypes.OPTIMIZE, StageTypes.WAIT): optimize_to_wait_transition,
        (StageTypes.OPTIMIZE, StageTypes.SETUP): invalid_transition,
        (StageTypes.OPTIMIZE, StageTypes.VALIDATE): invalid_transition,
        (StageTypes.OPTIMIZE, StageTypes.EXECUTE): optimize_to_execute_transition,
        (StageTypes.OPTIMIZE, StageTypes.TEARDOWN): invalid_transition,
        (StageTypes.OPTIMIZE, StageTypes.ANALYZE): invalid_transition,
        (StageTypes.OPTIMIZE, StageTypes.CHECKPOINT): optimize_to_checkpoint_transition,
        (StageTypes.OPTIMIZE, StageTypes.COMPLETE): invalid_transition,
        (StageTypes.OPTIMIZE, StageTypes.SUBMIT): invalid_transition,
        (StageTypes.OPTIMIZE, StageTypes.ERROR): error_transition,

        # State: Execute
        (StageTypes.EXECUTE, StageTypes.EXECUTE): execute_to_execute_transition,
        (StageTypes.EXECUTE, StageTypes.IDLE): invalid_transition,
        (StageTypes.EXECUTE, StageTypes.WAIT): execute_to_wait_transition,
        (StageTypes.EXECUTE, StageTypes.SETUP): execute_to_setup_transition,
        (StageTypes.EXECUTE, StageTypes.VALIDATE): invalid_transition,
        (StageTypes.EXECUTE, StageTypes.OPTIMIZE): execute_to_optimize_transition,
        (StageTypes.EXECUTE, StageTypes.TEARDOWN): execute_to_teardown_transition,
        (StageTypes.EXECUTE, StageTypes.ANALYZE): execute_to_analyze_transition,
        (StageTypes.EXECUTE, StageTypes.CHECKPOINT): execute_to_checkpoint_transition,
        (StageTypes.EXECUTE, StageTypes.SUBMIT): invalid_transition,
        (StageTypes.EXECUTE, StageTypes.COMPLETE): invalid_transition,
        (StageTypes.EXECUTE, StageTypes.ERROR): error_transition,

        # State: Teardown
        (StageTypes.TEARDOWN, StageTypes.TEARDOWN): invalid_transition,
        (StageTypes.TEARDOWN, StageTypes.IDLE): invalid_transition,
        (StageTypes.TEARDOWN, StageTypes.WAIT): teardown_to_wait_transition,
        (StageTypes.TEARDOWN, StageTypes.SETUP): invalid_transition,
        (StageTypes.TEARDOWN, StageTypes.VALIDATE): invalid_transition,
        (StageTypes.TEARDOWN, StageTypes.OPTIMIZE): invalid_transition,
        (StageTypes.TEARDOWN, StageTypes.EXECUTE): invalid_transition,
        (StageTypes.TEARDOWN, StageTypes.ANALYZE): teardown_to_analyze_transition,
        (StageTypes.TEARDOWN, StageTypes.CHECKPOINT): teardown_to_checkpoint_transition,
        (StageTypes.TEARDOWN, StageTypes.SUBMIT): invalid_transition,
        (StageTypes.TEARDOWN, StageTypes.COMPLETE): invalid_transition,
        (StageTypes.TEARDOWN, StageTypes.ERROR): error_transition,

        # State: Analyze
        (StageTypes.ANALYZE, StageTypes.ANALYZE): invalid_transition,
        (StageTypes.ANALYZE, StageTypes.IDLE): invalid_transition,
        (StageTypes.ANALYZE, StageTypes.WAIT): analyze_to_wait_transition,
        (StageTypes.ANALYZE, StageTypes.SETUP): invalid_transition,
        (StageTypes.ANALYZE, StageTypes.VALIDATE): invalid_transition,
        (StageTypes.ANALYZE, StageTypes.OPTIMIZE): invalid_transition,
        (StageTypes.ANALYZE, StageTypes.EXECUTE): invalid_transition,
        (StageTypes.ANALYZE, StageTypes.TEARDOWN): invalid_transition,
        (StageTypes.ANALYZE, StageTypes.CHECKPOINT): analyze_to_checkpoint_transition,
        (StageTypes.ANALYZE, StageTypes.SUBMIT): analyze_to_submit_transition,
        (StageTypes.ANALYZE, StageTypes.COMPLETE): invalid_transition,
        (StageTypes.ANALYZE, StageTypes.ERROR): error_transition,

        # State: Checkpoint
        (StageTypes.CHECKPOINT, StageTypes.CHECKPOINT): checkpoint_to_checkpoint_transition,
        (StageTypes.CHECKPOINT, StageTypes.IDLE): invalid_transition,
        (StageTypes.CHECKPOINT, StageTypes.WAIT): checkpoint_to_wait_transition,
        (StageTypes.CHECKPOINT, StageTypes.SETUP): checkpoint_to_setup_transition,
        (StageTypes.CHECKPOINT, StageTypes.VALIDATE): invalid_transition,
        (StageTypes.CHECKPOINT, StageTypes.OPTIMIZE): checkpoint_to_optimize_transition,
        (StageTypes.CHECKPOINT, StageTypes.EXECUTE): checkpoint_to_execute_transition,
        (StageTypes.CHECKPOINT, StageTypes.TEARDOWN): checkpoint_to_teardown_transition,
        (StageTypes.CHECKPOINT, StageTypes.ANALYZE): checkpoint_to_analyze_transition,
        (StageTypes.CHECKPOINT, StageTypes.SUBMIT): checkpoint_to_submit_transition,
        (StageTypes.CHECKPOINT, StageTypes.COMPLETE): checkpoint_to_complete_transition,
        (StageTypes.CHECKPOINT, StageTypes.ERROR): error_transition,


        # State: Submit
        (StageTypes.SUBMIT, StageTypes.SUBMIT): submit_to_submit_transition,
        (StageTypes.SUBMIT, StageTypes.IDLE): invalid_transition,
        (StageTypes.SUBMIT, StageTypes.WAIT): submit_to_wait_transition,
        (StageTypes.SUBMIT, StageTypes.SETUP): submit_to_setup_transition,
        (StageTypes.SUBMIT, StageTypes.VALIDATE): invalid_transition,
        (StageTypes.SUBMIT, StageTypes.OPTIMIZE): submit_to_optimize_transition,
        (StageTypes.SUBMIT, StageTypes.EXECUTE): submit_to_execute_transition,
        (StageTypes.SUBMIT, StageTypes.TEARDOWN): invalid_transition,
        (StageTypes.SUBMIT, StageTypes.ANALYZE): invalid_transition,
        (StageTypes.SUBMIT, StageTypes.CHECKPOINT): submit_to_checkpoint_transition,
        (StageTypes.SUBMIT, StageTypes.COMPLETE): submit_to_complete_transition,
        (StageTypes.SUBMIT, StageTypes.ERROR): error_transition,

        # State: Complete
        (StageTypes.COMPLETE, StageTypes.COMPLETE): exit_transition,
        (StageTypes.COMPLETE, StageTypes.IDLE): exit_transition,
        (StageTypes.COMPLETE, StageTypes.WAIT): exit_transition,
        (StageTypes.COMPLETE, StageTypes.SETUP): exit_transition,
        (StageTypes.COMPLETE, StageTypes.VALIDATE): exit_transition,
        (StageTypes.COMPLETE, StageTypes.OPTIMIZE): exit_transition,
        (StageTypes.COMPLETE, StageTypes.EXECUTE): exit_transition,
        (StageTypes.COMPLETE, StageTypes.TEARDOWN): exit_transition,
        (StageTypes.COMPLETE, StageTypes.ANALYZE): exit_transition,
        (StageTypes.COMPLETE, StageTypes.CHECKPOINT): exit_transition,
        (StageTypes.COMPLETE, StageTypes.SUBMIT): exit_transition,
        (StageTypes.COMPLETE, StageTypes.ERROR): error_transition,

        # State: Error
        (StageTypes.ERROR, StageTypes.ERROR): invalid_transition,
        (StageTypes.ERROR, StageTypes.IDLE): invalid_transition,
        (StageTypes.ERROR, StageTypes.WAIT): invalid_transition,
        (StageTypes.ERROR, StageTypes.SETUP): invalid_transition,
        (StageTypes.ERROR, StageTypes.VALIDATE): invalid_transition,
        (StageTypes.ERROR, StageTypes.OPTIMIZE): invalid_transition,
        (StageTypes.ERROR, StageTypes.EXECUTE): invalid_transition,
        (StageTypes.ERROR, StageTypes.TEARDOWN): invalid_transition,
        (StageTypes.ERROR, StageTypes.ANALYZE): invalid_transition,
        (StageTypes.ERROR, StageTypes.CHECKPOINT): invalid_transition,
        (StageTypes.ERROR, StageTypes.SUBMIT): invalid_transition,
        (StageTypes.ERROR, StageTypes.COMPLETE): invalid_transition
    }


from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.stages.base.stage import Stage
from hedra.logging import HedraLogger


async def wait_to_wait_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context
    return None, StageTypes.WAIT


async def wait_to_setup_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context
    return None, StageTypes.SETUP


async def wait_to_validate_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context
    return None, StageTypes.VALIDATE


async def wait_to_optimize_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context
    return None, StageTypes.OPTIMIZE


async def wait_to_execute_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context
    return None, StageTypes.EXECUTE


async def wait_to_teardown_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context
    return None, StageTypes.TEARDOWN


async def wait_to_analyze_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context
    return None, StageTypes.ANALYZE


async def wait_to_checkpoint_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context
    return None, StageTypes.CHECKPOINT


async def wait_to_submit_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context
    return None, StageTypes.SUBMIT


async def wait_to_complete_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context
    return None, StageTypes.COMPLETE


async def wait_to_error_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context 
    next_stage.context = current_stage.context
    return None, StageTypes.ERROR


from typing import Coroutine
from .types import PluginHooks


class PluginHook:

    def __init__(
        self, 
        name: str, 
        shortname: str,
        call: Coroutine, 
        plugin: str = None,
        hook_type=PluginHooks.CUSTOM,
    ) -> None:
        self.name = name
        self.shortname = shortname
        self.call = call
        self.plugin = plugin
        self.hook_type = hook_type


import asyncio
import uuid
import psutil
from concurrent.futures import ThreadPoolExecutor
from typing import Any, List
from numpy import float32, float64, int16, int32, int64
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet


try:
    import sqlalchemy
    from snowflake.sqlalchemy import URL
    from sqlalchemy import create_engine
    from sqlalchemy.schema import CreateTable
    from .snowflake_config import SnowflakeConfig
    has_connector = True

except Exception:
    snowflake = None
    SnowflakeConfig = None
    has_connector = False

class Snowflake:

    def __init__(self, config: SnowflakeConfig) -> None:
        self.username = config.username
        self.password = config.password
        self.organization_id = config.organization_id
        self.account_id = config.account_id
        self.private_key = config.private_key
        self.warehouse = config.warehouse
        self.database = config.database
        self.schema = config.database_schema

        self.events_table_name = config.events_table
        self.metrics_table_name = config.metrics_table
        self.shared_metrics_table_name = 'stage_metrics'
        self.errors_table_name = 'stage_errors'

        self.custom_fields = config.custom_fields
        self.connect_timeout = config.connect_timeout
        
        self.metadata = sqlalchemy.MetaData()
        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))

        self._engine = None
        self._connection = None
        self._events_table = None
        self._metrics_table = None
        self._shared_metrics_table = None
        self._custom_metrics_tables = {}
        self._errors_table = None

        self._loop = asyncio.get_event_loop()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to Snowflake instance at - Warehouse: {self.warehouse} - Database: {self.database} - Schema: {self.schema}')
            self._engine = await self._loop.run_in_executor(
                self._executor,
                create_engine,
                URL(
                    user=self.username,
                    password=self.password,
                    account=self.account_id,
                    warehouse=self.warehouse,
                    database=self.database,
                    schema=self.schema
                )
            
            )

            self._connection = await asyncio.wait_for(
                self._loop.run_in_executor(
                    self._executor,
                    self._engine.connect
                ),
                timeout=self.connect_timeout
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Snowflake instance at - Warehouse: {self.warehouse} - Database: {self.database} - Schema: {self.schema}')

        except asyncio.TimeoutError:
            raise Exception('Err. - Connection to Snowflake timed out - check your account id, username, and password.')


    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Table - {self.events_table_name}')
    
        for event in events:

            if self._events_table is None:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Events table - {self.events_table_name} - if not exists')
                events_table = sqlalchemy.Table(
                    self.events_table_name,
                    self.metadata,
                    sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                    sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('time', sqlalchemy.Float),
                    sqlalchemy.Column('succeeded', sqlalchemy.Boolean),
                )
                
                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    CreateTable(events_table, if_not_exists=True)
                )

                self._events_table = events_table
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Events table - {self.events_table_name}')
            
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._events_table.insert(values=event.record)
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Table - {self.events_table_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Table - {self.shared_metrics_table_name}')

        if self._shared_metrics_table is None:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Shared Metrics table - {self.shared_metrics_table_name} - if not exists')

            shared_metrics_table = sqlalchemy.Table(
                self.shared_metrics_table_name,
                self.metadata,
                sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                sqlalchemy.Column('group', sqlalchemy.TEXT),
                sqlalchemy.Column('total', sqlalchemy.BIGINT),
                sqlalchemy.Column('succeeded', sqlalchemy.BIGINT),
                sqlalchemy.Column('failed', sqlalchemy.BIGINT),
                sqlalchemy.Column('actions_per_second', sqlalchemy.FLOAT)
            )

            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                CreateTable(shared_metrics_table, if_not_exists=True)
            )

            self._shared_metrics_table = shared_metrics_table
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Shared Metrics table - {self.shared_metrics_table_name}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._shared_metrics_table.insert(values={
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'common',
                    **metrics_set.common_stats
                })
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Table - {self.shared_metrics_table_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Table - {self.metrics_table_name}')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            if self._metrics_table is None:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Metrics table - {self.metrics_table_name} - if not exists')

                metrics_table = sqlalchemy.Table(
                    self.metrics_table_name,
                    self.metadata,
                    sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                    sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('median', sqlalchemy.FLOAT),
                    sqlalchemy.Column('mean', sqlalchemy.FLOAT),
                    sqlalchemy.Column('variance', sqlalchemy.FLOAT),
                    sqlalchemy.Column('stdev', sqlalchemy.FLOAT),
                    sqlalchemy.Column('minimum', sqlalchemy.FLOAT),
                    sqlalchemy.Column('maximum', sqlalchemy.FLOAT)
                )

                for quantile in metrics_set.quantiles:
                    metrics_table.append_column(
                        sqlalchemy.Column(f'{quantile}', sqlalchemy.FLOAT)
                    )

                for custom_field_name, sql_alchemy_type in metrics_set.custom_schemas:
                    metrics_table.append_column(custom_field_name, sql_alchemy_type)  

                
                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    CreateTable(metrics_table, if_not_exists=True)
                )

                self._metrics_table = metrics_table
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Metrics table - {self.metrics_table_name}')

        for group_name, group in metrics_set.groups.items():
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._metrics_table.insert(values={
                    **group.record,
                    'group': group_name
                })
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Table - {self.metrics_table_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            for custom_group_name, group in metrics_set.custom_metrics.items():

                custom_table_name = f'{custom_group_name}_metrics'
                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to table - {custom_group_name}')

                if self._custom_metrics_tables.get(custom_table_name) is None:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Custom Metrics table - {custom_group_name} - if not exists')

                    custom_table = sqlalchemy.Table(
                        custom_table_name,
                        self.metadata,
                        sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                        sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('group', sqlalchemy.TEXT),
                    )

                    for field, value in group.items():

                        if isinstance(value, (int, int16, int32, int64)):
                            custom_table.append_column(
                                sqlalchemy.Column(field, sqlalchemy.Integer)
                            )
                        
                        elif isinstance(value, (float, float32, float64)):
                            custom_table.append_column(
                                sqlalchemy.Column(field, sqlalchemy.FLOAT)
                            )
                    
                    await self._loop.run_in_executor(
                        self._executor,
                        self._connection.execute,
                        CreateTable(custom_table, if_not_exists=True)
                    )

                    self._custom_metrics_tables[custom_table_name] = custom_table
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Custom Metrics table - {custom_group_name}')

                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    self._custom_metrics_tables[custom_table_name].insert(values={
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': custom_group_name,
                        **group
                    })
                )

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to table - {custom_group_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Table - {self.errors_table_name}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            if self._errors_table is None:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Error Metrics table - {self.errors_table_name} - if not exists')

                errors_table = sqlalchemy.Table(
                    self.errors_table_name,
                    self.metadata,
                    sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                    sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('stage', sqlalchemy.TEXT),
                    sqlalchemy.Column('error_message', sqlalchemy.TEXT),
                    sqlalchemy.Column('error_count', sqlalchemy.BIGINT)
                )

                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    CreateTable(errors_table, if_not_exists=True)
                )

                self._errors_table = errors_table
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Error Metrics table - {self.errors_table_name}')

            for error in metrics_set.errors:
                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    self._errors_table.insert(values={
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'error_message': error.get('message'),
                        'error_count': error.get('count')
                    })
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Table - {self.errors_table_name}')
        
    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connection to Snowflake instance at - Warehouse: {self.warehouse} - Database: {self.database} - Schema: {self.schema}')

        await self._loop.run_in_executor(
            None,
            self._connection.close
        )

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closed session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connection to Snowflake instance at - Warehouse: {self.warehouse} - Database: {self.database} - Schema: {self.schema}')


import asyncio
import uuid
from typing import Awaitable, Dict, List, Set, Tuple, Union
from hedra.core.engines.types.common import Timeouts
from hedra.core.engines.types.common.types import RequestTypes
from .context_config import ContextConfig
from .context_group import ContextGroup
from .pool import ContextPool
from .command import PlaywrightCommand
from .result import PlaywrightResult


PlaywrightResponseFuture = Awaitable[Union[PlaywrightResult, Exception]]
PlaywrightBatchResponseFuture = Awaitable[Tuple[Set[PlaywrightResponseFuture], Set[PlaywrightResponseFuture]]]


class MercuryPlaywrightClient:

    __slots__ = (
        'session_id',
        'pool',
        'timeouts',
        'registered',
        'closed',
        'config',
        'sem',
        'active',
        'waiter',
        '_discarded_context_groups',
        '_discarded_contexts',
        '_pending_context_groups',
        '_playwright_setup'
    )

    def __init__(self,  concurrency: int = 500, group_size: int=50, timeouts: Timeouts = Timeouts()) -> None:
        
        self.session_id = str(uuid.uuid4())

        self.pool = ContextPool(concurrency, group_size)
        self.timeouts = timeouts
        self.registered: Dict[str, PlaywrightCommand] = {}
        self.closed = False
        self.config = None

        self.sem = asyncio.Semaphore(value=concurrency)
        self.active = 0
        self.waiter = None

        self._discarded_context_groups: List[ContextGroup] = []
        self._discarded_contexts = []
        self._pending_context_groups: List[ContextGroup] = []
        self._playwright_setup = False
        

    async def setup(self, config: ContextConfig):
        if self._playwright_setup is False:
            self.config = config
            self.pool.create_pool(self.config)
            for context_group in self.pool:
                await context_group.create()

            self._playwright_setup = True

    async def prepare(self, command: PlaywrightCommand) -> Awaitable[None]:

        command.options.extra = {
            **command.options.extra,
            'timeout': self.timeouts.total_timeout * 1000
        }

        self.registered[command.name] = command

    def extend_pool(self, increased_capacity: int):
        self.pool.size += increased_capacity
        for _ in range(increased_capacity):
            context_group = ContextGroup(
                **self.config,
                concurrency=int(self.pool.size/self.pool.group_size)
            )

            self._pending_context_groups.append(context_group)

            self.pool.contexts.append(context_group)
        
        self.sem = asyncio.Semaphore(self.pool.size)

    def shrink_pool(self, decrease_capacity: int):
        self.pool.size -= decrease_capacity

        for context_group in self.pool.contexts[self.pool.size:]:
            self._discarded_context_groups.append(context_group)

        self.pool.contexts = self.pool.contexts[:self.pool.size]

        for context_group in self.pool.contexts:
            group_size = int(self.pool.size/self.pool.group_size)

            for context in context_group.contexts[group_size:]:
                self._discarded_contexts.append(context)

            context_group.contexts = context_group.contexts[:group_size]
            context_group.librarians = context_group.librarians[:group_size]

        self.sem = asyncio.Semaphore(self.pool.size)


    async def execute_prepared_command(self, command: PlaywrightCommand) -> PlaywrightResponseFuture:

        for pending_context in self._pending_context_groups:
            await pending_context.create()

        result = PlaywrightResult(command, type=RequestTypes.PLAYWRIGHT)
        self.active += 1
        
        async with self.sem:
            context = self.pool.contexts.pop()
            try:

                if command.hooks.listen:
                    event = asyncio.Event()
                    command.hooks.channel_events.append(event)
                    await event.wait()

                if command.hooks.before:
                    command = await command.hooks.before(command, result)

                result = await context.execute(command)

                if command.hooks.after:
                    command = await command.hooks.after(command, result)

                if command.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel(result, command.hooks.listeners)
                        ) for channel in command.hooks.channels
                    ])

                    for listener in command.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()  

                self.pool.contexts.append(context)

            except Exception as e:
                result.error = e
                self.pool.contexts.append(context)

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None 

            return result

    async def close(self):
        if self.closed is False:
            for context_group in self._discarded_context_groups:
                await context_group.close()

            for context in self._discarded_contexts:
                await context.close()

            self.closed = True
                    


# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
from .worker_pb2 import (
    LeaderUpdateRequest,
    WorkerServerUpdateResponse,
    NewJobRequest,
    NewJobResponse,
    PollWorkerRequest,
    PollWorkerResponse,
    JobCompleteRequest,
    JobCompleteResponse,
    WorkerHeartbeatRequest,
    WorkerHeartbeatResponse,
    LeaderRegisterWorkerRequest,
    LeaderRegisterWorkerResponse
)


class WorkerServerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetRequestsCompleted = channel.unary_unary(
                '/WorkerServer/GetRequestsCompleted',
                request_serializer=LeaderUpdateRequest.SerializeToString,
                response_deserializer=WorkerServerUpdateResponse.FromString,
                )
        self.CreateNewWorkerJob = channel.unary_unary(
                '/WorkerServer/CreateNewWorkerJob',
                request_serializer=NewJobRequest.SerializeToString,
                response_deserializer=NewJobResponse.FromString,
                )
        self.GetJobStatus = channel.unary_unary(
                '/WorkerServer/GetJobStatus',
                request_serializer=PollWorkerRequest.SerializeToString,
                response_deserializer=PollWorkerResponse.FromString,
                )
        self.GetJobResults = channel.unary_unary(
                '/WorkerServer/GetJobResults',
                request_serializer=JobCompleteRequest.SerializeToString,
                response_deserializer=JobCompleteResponse.FromString,
                )
        self.CheckWorkerHeartbeat = channel.unary_unary(
                '/WorkerServer/CheckWorkerHeartbeat',
                request_serializer=WorkerHeartbeatRequest.SerializeToString,
                response_deserializer=WorkerHeartbeatResponse.FromString,
                )
        self.LeaderRegisterWorker = channel.unary_unary(
                '/WorkerServer/LeaderRegisterWorker',
                request_serializer=LeaderRegisterWorkerRequest.SerializeToString,
                response_deserializer=LeaderRegisterWorkerResponse.FromString,
                )


class WorkerServerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetRequestsCompleted(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateNewWorkerJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobResults(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CheckWorkerHeartbeat(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LeaderRegisterWorker(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_WorkerServerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetRequestsCompleted': grpc.unary_unary_rpc_method_handler(
                    servicer.GetRequestsCompleted,
                    request_deserializer=LeaderUpdateRequest.FromString,
                    response_serializer=WorkerServerUpdateResponse.SerializeToString,
            ),
            'CreateNewWorkerJob': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateNewWorkerJob,
                    request_deserializer=NewJobRequest.FromString,
                    response_serializer=NewJobResponse.SerializeToString,
            ),
            'GetJobStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobStatus,
                    request_deserializer=PollWorkerRequest.FromString,
                    response_serializer=PollWorkerResponse.SerializeToString,
            ),
            'GetJobResults': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobResults,
                    request_deserializer=JobCompleteRequest.FromString,
                    response_serializer=JobCompleteResponse.SerializeToString,
            ),
            'CheckWorkerHeartbeat': grpc.unary_unary_rpc_method_handler(
                    servicer.CheckWorkerHeartbeat,
                    request_deserializer=WorkerHeartbeatRequest.FromString,
                    response_serializer=WorkerHeartbeatResponse.SerializeToString,
            ),
            'LeaderRegisterWorker': grpc.unary_unary_rpc_method_handler(
                    servicer.LeaderRegisterWorker,
                    request_deserializer=LeaderRegisterWorkerRequest.FromString,
                    response_serializer=LeaderRegisterWorkerResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'WorkerServer', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class WorkerServer(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetRequestsCompleted(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/WorkerServer/GetRequestsCompleted',
            LeaderUpdateRequest.SerializeToString,
            WorkerServerUpdateResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateNewWorkerJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/WorkerServer/CreateNewWorkerJob',
            NewJobRequest.SerializeToString,
            NewJobResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetJobStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/WorkerServer/GetJobStatus',
            PollWorkerRequest.SerializeToString,
            PollWorkerResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetJobResults(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/WorkerServer/GetJobResults',
            JobCompleteRequest.SerializeToString,
            JobCompleteResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CheckWorkerHeartbeat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/WorkerServer/CheckWorkerHeartbeat',
            WorkerHeartbeatRequest.SerializeToString,
            WorkerHeartbeatResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LeaderRegisterWorker(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/WorkerServer/LeaderRegisterWorker',
            LeaderRegisterWorkerRequest.SerializeToString,
            LeaderRegisterWorkerResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import asyncio
import datetime
from .timer import Timer


class ProgressText:
    def __init__(self):
        self.total_timer = Timer('Total')
        self.group_timer = Timer('Group')

        self.selected_timer_name = self.group_timer.name
        self.selected_timer = self.group_timer

        self.run_cli_task = False
        self.run_timer_task = False
        self.cli_message = ''
        self.cli_messages = []
        self.next_cli_message = 0
        self._cli_task = None
        self._timer_task = None
        self.enabled = True
        self.finalized = False
        self.group_finalized = False

    def __str__(self):

        self.total_timer.update()
        self.group_timer.update()

        if self.finalized:
            return f'{self.cli_message} - {self.total_timer.elapsed_message}'

        elif self.group_finalized:
            return f'{self.cli_message} - {self.group_timer.elapsed_message}'

        return f'> {self.cli_message} - {self.selected_timer.elapsed_message}'

    async def append_cli_message(self, text: str):
        self.cli_message = text
        self.cli_messages.append(text)
        await asyncio.sleep(1)

    def start_cli_tasks(self):
        if self.enabled:
            self.finalized = False
            self.group_finalized = False
            self._timer_task = asyncio.create_task(self._start_timer_tasks())
            self._cli_task = asyncio.create_task(self._start_cli_tasks())

    async def _start_cli_tasks(self):
        self.run_cli_task = True

        while self.run_cli_task:
            for cli_task in self.cli_messages:

                if self.run_cli_task is False:
                    return

                self.cli_message = cli_task
                await asyncio.sleep(2.5)

    async def _start_timer_tasks(self):
        self.run_timer_task = True
        self.total_timer.update()
        self.group_timer.update()

        while self.run_timer_task:
            if self.selected_timer_name == self.total_timer.name:
                self.selected_timer_name = self.group_timer.name
                self.selected_timer = self.group_timer

            else:
                self.selected_timer_name = self.total_timer.name
                self.selected_timer = self.total_timer

            await asyncio.sleep(2)
            if self.run_timer_task is False:
                return


    async def stop_cli_tasks(self):
        self.run_cli_task = False
        self.run_timer_task = False
        await self._cli_task
        await self._timer_task

        self.cli_messages = []

    async def pause_cli_tasks(self):
        self.run_cli_task = False
        if self._cli_task:
            await self._cli_task

    async def clear_and_replace(self, message: str):

        await self.pause_cli_tasks()
        self.cli_message = message
        self.cli_messages = [message]
        self.start_cli_tasks()



from .struct_types import (
    _STRUCT_B,
    _STRUCT_H,
    _STRUCT_HBBBL,
    _STRUCT_HL,
    _STRUCT_L,
    _STRUCT_LB,
    _STRUCT_LL
)
from .stream_associations import (
    _STREAM_ASSOC_EITHER,
    _STREAM_ASSOC_HAS_STREAM,
    _STREAM_ASSOC_NO_STREAM
)
from .frame_length import (
    FRAME_MAX_ALLOWED_LEN,
    FRAME_MAX_LEN
)
from .frame_flags import (
    Flag,
    Flags
)


from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.registry.registry_types import TaskHook
from hedra.core.graphs.stages.validate.exceptions import HookValidationError
from .base_hook_validator import BaseHookVaidator


class TaskHookValidator(BaseHookVaidator):

    def __init__(self, metadata_string: str) -> None:
        super(
            TaskHookValidator,
            self
        ).__init__(metadata_string)

    async def validate(self, hook: TaskHook):

        try:

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Validating {hook.hook_type.name.capitalize()} Hook - {hook.name}:{hook.hook_id}:{hook.hook_id}')

            assert hook.hook_type is HookType.TASK, f"Hook type mismatch - hook {hook.name}:{hook.hook_id} is a {hook.hook_type.name} hook, but Hedra expected a {HookType.TASK.name} hook."
            assert hook.shortname in hook.name, f"Shortname {hook.shortname} must be contained in full Hook name {hook.name}:{hook.hook_id} for @task hook {hook.name}:{hook.hook_id}."
            assert hook.call is not None, f"Method is not not found on stage or was not supplied to @task hook - {hook.name}:{hook.hook_id}"
            assert hook.call.__code__.co_argcount == 1, f"Too many args. - @task hook {hook.name}:{hook.hook_id} requires no additional args."
            assert 'self' in hook.call.__code__.co_varnames

            for notify_action in hook.notifiers:
                assert notify_action not in hook.listeners, f"Notify/listen loopback. - @task hook {hook.name}:{hook.hook_id} cannot notify and listen on same channel."

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Validated {hook.hook_type.name.capitalize()} Hook - {hook.name}:{hook.hook_id}:{hook.hook_id}')


        except AssertionError as hook_validation_error:
                raise HookValidationError(hook.stage_instance, str(hook_validation_error))


import asyncio
import functools
import psutil
import uuid
from typing import List
from numpy import float32, float64, int16, int32, int64
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet

try:
    from google.cloud import bigquery
    from .bigquery_config import BigQueryConfig
    has_connector = True

except Exception:
    bigquery = None
    BigQueryConfig = None
    has_connector = False


class BigQuery:

    def __init__(self, config: BigQueryConfig) -> None:
        self.service_account_json_path = config.service_account_json_path
        self.project_name = config.project_name
        self.dataset_name = config.dataset_name
        self.retry_timeout = config.retry_timeout
        self.events_table_name = config.events_table
        self.metrics_table_name = config.metrics_table
        self.shared_metrics_table_name = f'{self.metrics_table_name}_shared'
        self.custom_metrics_table_names = {}
        self.errors_table_name = f'{self.metrics_table_name}_errors'
        self.custom_fields = config.custom_fields or []

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self.credentials = None
        self.client = None
        self._events_table = None
        self._errors_table = None
        self._metrics_table = None
        self._custom_metrics_tables = {}
        self._shared_metrics_table = None

        self._loop = asyncio.get_event_loop()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to Google Cloud - Loading account config from - {self.service_account_json_path}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Opening session - {self.session_uuid}')

        self.client = bigquery.Client.from_service_account_json(self.service_account_json_path)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opened connection to Google Cloud - Loaded account config from - {self.service_account_json_path}')

    async def submit_events(self, events: List[BaseEvent]):
        
        if self._events_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.events_table_name} - if not exists')
            
            events_table_name = f'{self.project_name}.{self.dataset_name}.{self.events_table_name}'

            table_schema = bigquery.Table(
                events_table_name,
                schema=[
                    bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('stage', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('time', 'FLOAT', mode='REQUIRED'),
                    bigquery.SchemaField('succeeded', 'BOOLEAN', mode='REQUIRED'),
                    bigquery.SchemaField('error', 'STRING')
                ])

            table_reference = bigquery.TableReference(
                bigquery.DatasetReference(
                    self.project_name,
                    self.dataset_name
                ),
                self.events_table_name
            )

            self._events_table = bigquery.Table(
                table_reference,
                schema=table_schema
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_table,
                    self._events_table,
                    exists_ok=True
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.events_table_name} - if not exists')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.events_table_name}')

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                table_reference,
                [event.record for event in events],
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.events_table_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):
        if self.shared_metrics_table_name is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.shared_metrics_table_name} - if not exists')

            table_schema = [
                bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                bigquery.SchemaField('stage', 'STRING', mode='REQUIRED'),
                bigquery.SchemaField('group', 'STRING', mode='REQUIRED'),
                bigquery.SchemaField('total', 'INTEGER', mode='REQUIRED'),
                bigquery.SchemaField('succeeded', 'INTEGER', mode='REQUIRED'),
                bigquery.SchemaField('failed', 'INTEGER', mode='REQUIRED'),
                bigquery.SchemaField('actions_per_second', 'FLOAT', mode='REQUIRED')
            ]

            table_reference = bigquery.TableReference(
                bigquery.DatasetReference(
                    self.project_name,
                    self.dataset_name
                ),
                self.shared_metrics_table_name
            )

            shared_metrics_table = bigquery.Table(
                table_reference,
                schema=table_schema
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_table,
                    shared_metrics_table,
                    exists_ok=True
                )
            )

            self._shared_metrics_table = shared_metrics_table

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.shared_metrics_table_name} - if not exists')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.shared_metrics_table_name} - if not exists')

        rows = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            rows.append({
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'group': 'common',
                **metrics_set.common_stats
            })

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                self._shared_metrics_table,
                rows,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.shared_metrics_table_name} - if not exists')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.metrics_table_name} - if not exists')

        rows = []
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            if self._metrics_table is None:

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.metrics_table_name} - if not exists')

                table_schema = [
                    bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('stage', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('group', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('median', 'FLOAT', mode='REQUIRED'),
                    bigquery.SchemaField('mean', 'FLOAT', mode='REQUIRED'),
                    bigquery.SchemaField('variance', 'FLOAT', mode='REQUIRED'),
                    bigquery.SchemaField('stdev','FLOAT', mode='REQUIRED'),
                    bigquery.SchemaField('minimum', 'FLOAT', mode='REQUIRED'),
                    bigquery.SchemaField('maximum', 'FLOAT', mode='REQUIRED')
                ]

                for quantile in metrics_set.quantiles:
                    table_schema.append(
                        bigquery.SchemaField(f'{quantile}', 'FLOAT')
                    )

                for custom_field_name, bigquery_schema_field in metrics_set.custom_schemas.items():
                    table_schema.append(
                        custom_field_name, 
                        bigquery_schema_field
                    ) 

                table_reference = bigquery.TableReference(
                    bigquery.DatasetReference(
                        self.project_name,
                        self.dataset_name
                    ),
                    self.metrics_table_name
                )

                metrics_table = bigquery.Table(
                    table_reference,
                    schema=table_schema
                )

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.create_table,
                        metrics_table,
                        exists_ok=True
                    )
                )

                self._metrics_table = metrics_table

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.metrics_table_name} - if not exists')
            
            for group_name, group in metrics_set.groups.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {group_name}:{group.metrics_group_id}')
                rows.append({
                    'group': group_name,
                    **group.record
                })

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                self._metrics_table,
                rows,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.metrics_table_name} - if not exists')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics - Project: {self.project_name} - Dataset: {self.dataset_name}')

        rows = defaultdict(list)
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            for custom_metrics_group_name, custom_metrics_group in metrics_set.custom_metrics.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Group - {custom_metrics_group_name}')

                custom_metrics_table_name = f'{custom_metrics_group_name}_metrics'

                if self._custom_metrics_tables.get(custom_metrics_table_name) is None:

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {custom_metrics_table_name} - if not exists')

                    table_schema = [
                        bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                        bigquery.SchemaField('stage', 'STRING', mode='REQUIRED'),
                        bigquery.SchemaField('group', 'STRING', mode='REQUIRED')
                    ]

                    for field, value in custom_metrics_group.items():

                        if isinstance(value, (int, int16, int32, int64)):
                            table_schema.append(
                                bigquery.SchemaField(field, 'INTEGER', mode='REQUIRED')
                            )

                        elif isinstance(value, (float, float32, float64)):
                            table_schema.append(
                                bigquery.SchemaField(field, 'FLOAT', mode='REQUIRED')
                            )

                    table_reference = bigquery.TableReference(
                        bigquery.DatasetReference(
                            self.project_name,
                            self.dataset_name
                        ),
                        custom_metrics_table_name
                    )

                    custom_metrics_table = bigquery.Table(
                        table_reference,
                        schema=table_schema
                    )

                    await self._loop.run_in_executor(
                        self._executor,
                        functools.partial(
                            self.client.create_table,
                            custom_metrics_table,
                            exists_ok=True
                        )
                    )

                    self._custom_metrics_tables[custom_metrics_table_name] = custom_metrics_table

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {custom_metrics_table_name} - if not exists')
            
                rows[custom_metrics_table_name].append({
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': custom_metrics_group_name,
                    **custom_metrics_group
                })
        
        for table_name, rows in rows.items():

            table = self._custom_metrics_tables.get(table_name)

            await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                table,
                rows,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics - Project: {self.project_name} - Dataset: {self.dataset_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Errors Metrics Set - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.errors_table_name} - if not exists')

        rows = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Errors Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            if self._errors_table is None:

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.errors_table_name} - if not exists')

                table_schema = [
                    bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('stage', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('error_message', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('error_count', 'INTEGER', mode='REQUIRED'),
                ]

                table_reference = bigquery.TableReference(
                    bigquery.DatasetReference(
                        self.project_name,
                        self.dataset_name
                    ),
                    self.errors_table_name
                )

                errors_table = bigquery.Table(
                    table_reference,
                    schema=table_schema
                )

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.create_table,
                        errors_table,
                        exists_ok=True
                    )
                )

                self._errors_table = errors_table

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.errors_table_name} - if not exists')

            rows.extend([
                {
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'error_message': error.get('message'),
                    'error_count': error.get('count')
                } for error in metrics_set.errors
            ])

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                self._errors_table,
                rows,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.errors_table_name} - if not exists')

    async def close(self):

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')

        await self._loop.run_in_executor(
            self._executor,
            self.client.close
        )